@@ -25,8 +25,6 @@
 #define BAD_HEADER	(-1)
 #define NO_ERROR	(0)
 
-#define XLOG_SET(f,b)	(((f) & (b)) == (b))
-
 static int logBBsize;
 char *trans_type[] = {
 	"",
@@ -71,6 +69,7 @@ char *trans_type[] = {
 	"GROWFSRT_FREE",
 	"SWAPEXT",
 	"SB_COUNT",
+	"CHECKPOINT",
 };
 
 typedef struct xlog_split_item {
@@ -306,12 +305,14 @@ xlog_print_trans_buffer(xfs_caddr_t *ptr, int len, int *i, int num_ops)
 			 */
 			memmove(&x, *ptr, sizeof(__be64));
 			memmove(&y, *ptr+8, sizeof(__be64));
-			printf(_("icount: %lld  ifree: %lld  "),
-				be64_to_cpu(x), be64_to_cpu(y));
+		       printf(_("icount: %llu  ifree: %llu  "),
+			       (unsigned long long) be64_to_cpu(x),
+			       (unsigned long long) be64_to_cpu(y));
 			memmove(&x, *ptr+16, sizeof(__be64));
 			memmove(&y, *ptr+24, sizeof(__be64));
-			printf(_("fdblks: %lld  frext: %lld\n"),
-				be64_to_cpu(x), be64_to_cpu(y));
+		       printf(_("fdblks: %llu  frext: %llu\n"),
+			       (unsigned long long) be64_to_cpu(x),
+			       (unsigned long long) be64_to_cpu(y));
 		}
 		super_block = 0;
 	} else if (be32_to_cpu(*(__be32 *)(*ptr)) == XFS_AGI_MAGIC) {
@@ -394,18 +395,22 @@ xlog_print_trans_buffer(xfs_caddr_t *ptr, int len, int *i, int num_ops)
 				dq->d_version, dq->d_flags,
 				be32_to_cpu(dq->d_id));
 			printf(_("blk limits  hard: %llu  soft: %llu\n"),
-				be64_to_cpu(dq->d_blk_hardlimit),
-				be64_to_cpu(dq->d_blk_softlimit));
+			       (unsigned long long)
+				       be64_to_cpu(dq->d_blk_hardlimit),
+			       (unsigned long long)
+				       be64_to_cpu(dq->d_blk_softlimit));
 			printf(_("blk  count: %llu  warns: %d  timer: %d\n"),
-				be64_to_cpu(dq->d_bcount),
-				be16_to_cpu(dq->d_bwarns),
+			       (unsigned long long) be64_to_cpu(dq->d_bcount),
+			       (int) be16_to_cpu(dq->d_bwarns),
 				be32_to_cpu(dq->d_btimer));
 			printf(_("ino limits  hard: %llu  soft: %llu\n"),
-				be64_to_cpu(dq->d_ino_hardlimit),
-				be64_to_cpu(dq->d_ino_softlimit));
+			       (unsigned long long)
+				       be64_to_cpu(dq->d_ino_hardlimit),
+			       (unsigned long long)
+				       be64_to_cpu(dq->d_ino_softlimit));
 			printf(_("ino  count: %llu  warns: %d  timer: %d\n"),
-				be64_to_cpu(dq->d_icount),
-				be16_to_cpu(dq->d_iwarns),
+			       (unsigned long long) be64_to_cpu(dq->d_icount),
+			       (int) be16_to_cpu(dq->d_iwarns),
 				be32_to_cpu(dq->d_itimer));
 		}
 	} else {
@@ -573,7 +578,7 @@ xlog_print_dir_sf(xfs_dir_shortform_t *sfp, int size)
 	printf(_("SHORTFORM DIRECTORY size %d count %d\n"),
 	       size, sfp->hdr.count);
 	memmove(&ino, &(sfp->hdr.parent), sizeof(ino));
-	printf(_(".. ino 0x%llx\n"), be64_to_cpu(*(__be64 *)&ino));
+       printf(_(".. ino 0x%llx\n"), (unsigned long long) be64_to_cpu(ino));
 
 	count = (uint)(sfp->hdr.count);
 	sfep = &(sfp->list[0]);
@@ -588,7 +593,11 @@ xlog_print_dir_sf(xfs_dir_shortform_t *sfp, int size)
 }
 
 int
-xlog_print_trans_inode(xfs_caddr_t *ptr, int len, int *i, int num_ops)
+xlog_print_trans_inode(xfs_caddr_t *ptr,
+		       int len,
+		       int *i,
+		       int num_ops,
+		       boolean_t continued)
 {
     xfs_icdinode_t	   dino;
     xlog_op_header_t	   *op_head;
@@ -610,8 +619,9 @@ xlog_print_trans_inode(xfs_caddr_t *ptr, int len, int *i, int num_ops)
     memmove(&src_lbuf, *ptr, MIN(sizeof(xfs_inode_log_format_64_t), len));
     (*i)++;					/* bump index */
     *ptr += len;
-    if (len == sizeof(xfs_inode_log_format_32_t) ||
-	len == sizeof(xfs_inode_log_format_64_t)) {
+    if (!continued &&
+	(len == sizeof(xfs_inode_log_format_32_t) ||
+	 len == sizeof(xfs_inode_log_format_64_t))) {
 	f = xfs_inode_item_format_convert((char*)&src_lbuf, len, &dst_lbuf);
 	printf(_("INODE: "));
 	printf(_("#regs: %d   ino: 0x%llx  flags: 0x%x   dsize: %d\n"),
@@ -634,7 +644,7 @@ xlog_print_trans_inode(xfs_caddr_t *ptr, int len, int *i, int num_ops)
     op_head = (xlog_op_header_t *)*ptr;
     xlog_print_op_header(op_head, *i, ptr);
 
-    if (XLOG_SET(op_head->oh_flags, XLOG_CONTINUE_TRANS))  {
+    if (op_head->oh_flags & XLOG_CONTINUE_TRANS)  {
 	return f->ilf_size-1;
     }
 
@@ -642,7 +652,7 @@ xlog_print_trans_inode(xfs_caddr_t *ptr, int len, int *i, int num_ops)
     mode = dino.di_mode & S_IFMT;
     size = (int)dino.di_size;
     xlog_print_trans_inode_core(&dino);
-    *ptr += sizeof(xfs_dinode_core_t);
+    *ptr += sizeof(xfs_icdinode_t);
 
     if (*i == num_ops-1 && f->ilf_size == 3)  {
 	return 1;
@@ -650,97 +660,76 @@ xlog_print_trans_inode(xfs_caddr_t *ptr, int len, int *i, int num_ops)
 
     /* does anything come next */
     op_head = (xlog_op_header_t *)*ptr;
-    switch (f->ilf_fields & XFS_ILOG_NONCORE) {
-	case XFS_ILOG_DEXT: {
-	    ASSERT(f->ilf_size == 3);
-	    (*i)++;
-	    xlog_print_op_header(op_head, *i, ptr);
-	    printf(_("EXTENTS inode data\n"));
-	    *ptr += be32_to_cpu(op_head->oh_len);
-	    if (XLOG_SET(op_head->oh_flags, XLOG_CONTINUE_TRANS))  {
-		return 1;
-	    }
-	    break;
-	}
-	case XFS_ILOG_DBROOT: {
-	    ASSERT(f->ilf_size == 3);
-	    (*i)++;
-	    xlog_print_op_header(op_head, *i, ptr);
-	    printf(_("BTREE inode data\n"));
-	    *ptr += be32_to_cpu(op_head->oh_len);
-	    if (XLOG_SET(op_head->oh_flags, XLOG_CONTINUE_TRANS))  {
-		return 1;
-	    }
-	    break;
-	}
-	case XFS_ILOG_DDATA: {
-	    ASSERT(f->ilf_size == 3);
-	    (*i)++;
-	    xlog_print_op_header(op_head, *i, ptr);
-	    printf(_("LOCAL inode data\n"));
-	    if (mode == S_IFDIR) {
-		xlog_print_dir_sf((xfs_dir_shortform_t*)*ptr, size);
-	    }
-	    *ptr += be32_to_cpu(op_head->oh_len);
-	    if (XLOG_SET(op_head->oh_flags, XLOG_CONTINUE_TRANS)) {
-		return 1;
-	    }
-	    break;
-	}
-	case XFS_ILOG_AEXT: {
-	    ASSERT(f->ilf_size == 3);
-	    (*i)++;
-	    xlog_print_op_header(op_head, *i, ptr);
-	    printf(_("EXTENTS inode attr\n"));
-	    *ptr += be32_to_cpu(op_head->oh_len);
-	    if (XLOG_SET(op_head->oh_flags, XLOG_CONTINUE_TRANS))  {
-		return 1;
-	    }
-	    break;
-	}
-	case XFS_ILOG_ABROOT: {
-	    ASSERT(f->ilf_size == 3);
-	    (*i)++;
-	    xlog_print_op_header(op_head, *i, ptr);
-	    printf(_("BTREE inode attr\n"));
-	    *ptr += be32_to_cpu(op_head->oh_len);
-	    if (XLOG_SET(op_head->oh_flags, XLOG_CONTINUE_TRANS))  {
-		return 1;
-	    }
-	    break;
-	}
-	case XFS_ILOG_ADATA: {
-	    ASSERT(f->ilf_size == 3);
-	    (*i)++;
-	    xlog_print_op_header(op_head, *i, ptr);
-	    printf(_("LOCAL inode attr\n"));
-	    if (mode == S_IFDIR) {
-		xlog_print_dir_sf((xfs_dir_shortform_t*)*ptr, size);
-	    }
-	    *ptr += be32_to_cpu(op_head->oh_len);
-	    if (XLOG_SET(op_head->oh_flags, XLOG_CONTINUE_TRANS)) {
-		return 1;
-	    }
-	    break;
-	}
-	case XFS_ILOG_DEV: {
-	    ASSERT(f->ilf_size == 2);
-	    printf(_("DEV inode: no extra region\n"));
-	    break;
-	}
-	case XFS_ILOG_UUID: {
-	    ASSERT(f->ilf_size == 2);
-	    printf(_("UUID inode: no extra region\n"));
-	    break;
-	}
-	case 0: {
-	    ASSERT(f->ilf_size == 2);
-	    break;
-	}
-	default: {
-	    xlog_panic(_("xlog_print_trans_inode: illegal inode type"));
-	}
+
+    switch (f->ilf_fields & (XFS_ILOG_DEV | XFS_ILOG_UUID)) {
+    case XFS_ILOG_DEV:
+	printf(_("DEV inode: no extra region\n"));
+	break;
+    case XFS_ILOG_UUID:
+	printf(_("UUID inode: no extra region\n"));
+	break;
     }
+
+    /* Only the inode core is logged */
+    if (f->ilf_size == 2)
+	return 0;
+
+    ASSERT(f->ilf_size <= 4);
+    ASSERT((f->ilf_size == 3) || (f->ilf_fields & XFS_ILOG_AFORK));
+
+    if (f->ilf_fields & XFS_ILOG_DFORK) {
+	    (*i)++;
+	    xlog_print_op_header(op_head, *i, ptr);
+
+	    switch (f->ilf_fields & XFS_ILOG_DFORK) {
+	    case XFS_ILOG_DEXT:
+		printf(_("EXTENTS inode data\n"));
+		break;
+	    case XFS_ILOG_DBROOT:
+		printf(_("BTREE inode data\n"));
+		break;
+	    case XFS_ILOG_DDATA:
+		printf(_("LOCAL inode data\n"));
+		if (mode == S_IFDIR)
+		    xlog_print_dir_sf((xfs_dir_shortform_t*)*ptr, size);
+		break;
+	    default:
+		ASSERT((f->ilf_fields & XFS_ILOG_DFORK) == 0);
+		break;
+	    }
+
+	    *ptr += be32_to_cpu(op_head->oh_len);
+	    if (op_head->oh_flags & XLOG_CONTINUE_TRANS)
+		return 1;
+	    op_head = (xlog_op_header_t *)*ptr;
+    }
+
+    if (f->ilf_fields & XFS_ILOG_AFORK) {
+	    (*i)++;
+	    xlog_print_op_header(op_head, *i, ptr);
+
+	    switch (f->ilf_fields & XFS_ILOG_AFORK) {
+	    case XFS_ILOG_AEXT:
+		printf(_("EXTENTS attr data\n"));
+		break;
+	    case XFS_ILOG_ABROOT:
+		printf(_("BTREE attr data\n"));
+		break;
+	    case XFS_ILOG_ADATA:
+		printf(_("LOCAL attr data\n"));
+		if (mode == S_IFDIR)
+		    xlog_print_dir_sf((xfs_dir_shortform_t*)*ptr, size);
+		break;
+	    default:
+		ASSERT((f->ilf_fields & XFS_ILOG_AFORK) == 0);
+		break;
+	    }
+	    *ptr += be32_to_cpu(op_head->oh_len);
+	    if (op_head->oh_flags & XLOG_CONTINUE_TRANS)
+		return 1;
+	    op_head = (xlog_op_header_t *)*ptr;
+    }
+
     return 0;
 }	/* xlog_print_trans_inode */
 
@@ -938,16 +927,18 @@ xlog_print_record(int			  fd,
 
     ptr = buf;
     for (i=0; i<num_ops; i++) {
+	int continued;
+
 	xlog_op_header_t *op_head = (xlog_op_header_t *)ptr;
 
 	print_xlog_op_line();
 	xlog_print_op_header(op_head, i, &ptr);
+	continued = ((op_head->oh_flags & XLOG_WAS_CONT_TRANS) ||
+		     (op_head->oh_flags & XLOG_CONTINUE_TRANS));
 
 	/* print transaction data */
 	if (print_no_data ||
-	    ((XLOG_SET(op_head->oh_flags, XLOG_WAS_CONT_TRANS) ||
-	      XLOG_SET(op_head->oh_flags, XLOG_CONTINUE_TRANS)) &&
-	     be32_to_cpu(op_head->oh_len) == 0)) {
+	    (continued && be32_to_cpu(op_head->oh_len) == 0)) {
 	    for (n = 0; n < be32_to_cpu(op_head->oh_len); n++) {
 		printf("%c", *ptr);
 		ptr++;
@@ -976,7 +967,7 @@ xlog_print_record(int			  fd,
 		    case XFS_LI_INODE: {
 			skip = xlog_print_trans_inode(&ptr,
 					be32_to_cpu(op_head->oh_len),
-					&i, num_ops);
+					&i, num_ops, continued);
 			break;
 		    }
 		    case XFS_LI_DQUOT: {
